uniform sampler2D dotexture;
uniform sampler2D ndtexture;
uniform sampler2D bouncetexture;
uniform sampler2D dpstexture;

uniform vec2 viewport;

uniform int blurSize;
uniform bool dps;
uniform int shadowBlurSize;

vec2 textureCoord(const vec2 coord){
	return coord/viewport;
}

vec4 getFrag(in vec2 coord, in sampler2D tex){
	return texture( tex, coord );
}

vec4 getND(in vec2 coord){
	return texture( ndtexture, coord );
}

vec4 getDO(in vec2 coord){
	return texture( dotexture, coord );
}

vec4 getBounce(in vec2 coord){
	return texture( bouncetexture, coord );
}

vec4 getDps(in vec2 coord){
	return texture( dpstexture, coord );
}

//////////////////////////////////////////////////

vec4 getBilateralBlur(in vec2 currCoord, in int size, in sampler2D texture)
{	
	int iter = size+1;
	int halfSize = int(floor(size/2.0));
	float sizef = float(size);

	vec4 sample;
	vec4 currFrag, sampleFrag;
	vec3 currNorm, sampleNorm;
	float currDepth, sampleDepth;

	currFrag = getND(currCoord);
	currNorm = currFrag.xyz;
	currDepth = currFrag.w;

	float texelSize = 1.0 / viewport.y;

	float normalWeight, depthWeight, weight, spacialWeight;
	float weightSum = 0.0;
	vec4 blur = vec4(0.0);

	vec2 sampleCoord;
	for (float y=-halfSize; y<=halfSize; y++) {
		
		sampleCoord = currCoord + vec2( 0.0, y*texelSize);
		sample = getFrag( sampleCoord, texture );
		
		sampleFrag = getND(sampleCoord);
		sampleNorm = sampleFrag.xyz;
		sampleDepth = sampleFrag.w;

		if(sampleDepth<1.0){
			normalWeight = max(0.0, dot(currNorm,sampleNorm));
			depthWeight = 1.0 / (1.0 + abs(currDepth - sampleDepth));
			spacialWeight = 1.0 - abs(y)/sizef;
		
			weight = spacialWeight * pow(normalWeight,4.0) * pow(depthWeight,8.0);

			weightSum += weight;
			blur += sample * weight;
		}
	}
	return blur/weightSum;
}

void getBilateralBlurBoth(in vec2 currCoord, in int size, inout vec4 doBlur, inout vec4 bounceBlur)
{	
	int iter = size+1;
	int halfSize = int(floor(size/2.0));
	float sizef = float(size);

	vec4 sampleDO, sampleBounce;
	vec4 currFrag, sampleFrag;
	vec3 currNorm, sampleNorm;
	float currDepth, sampleDepth;
	
	currFrag = getND(currCoord);
	currNorm = currFrag.xyz;
	currDepth = currFrag.w;
	
	float texelSize = 1.0 / viewport.y;

	float normalWeight, depthWeight, weight, spacialWeight;
	float weightSum = 0.0;

	vec2 sampleCoord;
	for (float y=-halfSize; y<=halfSize; y++) {

		sampleCoord = currCoord + vec2( 0.0, y*texelSize);
		
		sampleDO = getFrag( sampleCoord, dotexture );
		sampleBounce = getFrag( sampleCoord, bouncetexture );
		
		sampleFrag = getND(sampleCoord);
		sampleNorm = sampleFrag.xyz;
		sampleDepth = sampleFrag.w;

		if(sampleDepth<1.0){
			normalWeight = max(0.0, dot(currNorm,sampleNorm));
			depthWeight = 1.0 / (1.0 + abs(currDepth - sampleDepth));
			spacialWeight = 1.0 - abs(y)/sizef;

			weight = spacialWeight * pow(normalWeight,4.0) * pow(depthWeight,8.0);
		
			weightSum += weight;
			doBlur += sampleDO * weight;	
			bounceBlur += sampleBounce * weight;
		}
	}
	doBlur /= weightSum;
	bounceBlur /= weightSum;
}

//////////////////////////////////////////////////////

void main( void )
{
	vec2 fragCoord = textureCoord(gl_FragCoord.xy);
	
	if (blurSize>0)
	{	
		vec4 doBlur = vec4(0.0);
		vec4 bounceBlur = vec4(0.0);
		getBilateralBlurBoth(fragCoord, blurSize, doBlur, bounceBlur);

		gl_FragData[0] = doBlur;
		gl_FragData[2] = bounceBlur;
	}else{
		gl_FragData[0] = getDO(fragCoord);
		gl_FragData[2] = getBounce(fragCoord);
	}

	if (shadowBlurSize>0 && dps)
	{
		vec4 dpsBlur = vec4(0.0);
		dpsBlur = getBilateralBlur(fragCoord, shadowBlurSize, dpstexture);
		gl_FragData[1] = dpsBlur;
	}else{
		gl_FragData[1] = getDps(fragCoord);
	}
}