uniform sampler2D ndtexture;
uniform sampler2D ndtextureback;
uniform sampler2D colortexture;
uniform sampler2D sampleTexture;
uniform sampler2D dpstexture;
uniform sampler2D lighttexture;

uniform vec2 viewport;
uniform vec2 giTextureSize;
uniform mat4 farPlaneVertices;
uniform vec3 cameraPos;

uniform int sampleCountB;
uniform float rmax;

mat3 toTextureMatrix;
mat3 toTextureMatrixInverse;

float near, far;
vec3 fnright, fnup, fneye;
float fwidth, fheight;

////////////////////////////////////////////////////////////////////////////////////

vec2 textureCoord(const vec2 coord){
	return coord/giTextureSize;
}

vec3 getNormal(in vec2 coord){
	return texture( ndtexture, coord ).xyz;
}

vec3 getNormalBack(in vec2 coord){
	return texture( ndtextureback, coord ).xyz;
}


float getDepth(in vec2 coord){
	return texture( ndtexture, coord ).a;
}

float getDepthBack(in vec2 coord){
	return texture( ndtextureback, coord ).a;
}

float getDpsMask(in vec2 coord){
	return texture( dpstexture, coord ).w;
}

vec3 getColor(in vec2 coord){	
	return texture( colortexture, coord ).xyz;
}

vec3 getLight(in vec2 coord){
	return texture( lighttexture, coord ).xyz;
}

vec3 getSample(in int i, in float j)
{	
	vec2 coord;
	coord.x = float(i)/float(sampleCountB);
	coord.y = j/256.0;
	return 2.0*texture( sampleTexture, coord ).xyz - 1.0;
}

mat3 inverseMatrix(in mat3 m)
{
	float det00, det01, det02, det10, det11, det12, det20, det21, det22;

	float detA = determinant(m);

	det00 = determinant(mat2( m[1].y, m[2].y, m[1].z, m[2].z ));
	det01 = determinant(mat2( m[0].z, m[2].z, m[0].y, m[2].y ));
	det02 = determinant(mat2( m[0].y, m[1].y, m[0].z, m[1].z ));

	det10 = determinant(mat2( m[1].z, m[2].z, m[1].x, m[2].x ));
	det11 = determinant(mat2( m[0].x, m[2].x, m[0].z, m[2].z ));
	det12 = determinant(mat2( m[0].z, m[1].z, m[0].x, m[1].x ));

	det20 = determinant(mat2( m[1].x, m[2].x, m[1].y, m[2].y ));
	det21 = determinant(mat2( m[0].y, m[2].y, m[0].x, m[2].x ));
	det22 = determinant(mat2( m[0].x, m[1].x, m[0].y, m[1].y ));

	return mat3(det00, det01, det02, det10, det11, det12, det20, det21, det22)/detA;
}

void setRealWorld()
{
	vec3 fbl, fbr, fur, ful, fcenter;
	vec3 fright, fup, feye;

	float znear = gl_ProjectionMatrix[2].z;
	float zfar = gl_ProjectionMatrix[3].z;
	near = zfar / (znear - 1.0);
	far = zfar / (znear + 1.0);

	fbl = farPlaneVertices[0].xyz;
	fbr = farPlaneVertices[1].xyz;
	fur = farPlaneVertices[2].xyz;
	ful = farPlaneVertices[3].xyz;

	fcenter = (fbl+fbr+ful+fur)*0.25;

	fright = fbr - fbl;
	fup = ful - fbl;
	feye = fcenter-cameraPos;
	fwidth = length(fright);
	fheight = length(fup);

	fnright = normalize(fright);
	fnup = normalize(fup);
	fneye = normalize(feye);

	toTextureMatrix = transpose(mat3(fnright, fnup, fneye));
	toTextureMatrixInverse = inverseMatrix( toTextureMatrix );
}

float viewDepth(in vec3 pos)
{	
	vec3 viewVec = pos-cameraPos;
	float depth = length(viewVec);
	float angle = dot(normalize(viewVec),fneye);

	return depth*angle/far;
}

vec3 getWorldPos(const vec2 coord, const float depth)
{
	vec3 tcoord = vec3( (coord-0.5)*vec2(fwidth,fheight), far);
	vec3 viewVec = toTextureMatrixInverse*tcoord;
	
	vec3 worldPos = cameraPos + depth * viewVec;
    return worldPos;
}

vec2 getTexturePos(const vec3 wpos, const float depth)
{
	vec3 viewVec = ( wpos - cameraPos ) / depth;
	vec3 coord = toTextureMatrix*viewVec;

	return coord.xy / vec2(fwidth,fheight) + 0.5;
}

////////////////////////////////////////////////////////////////////////////////////

vec3 getBounce(in vec2 currCoord)
{
	vec2 occlCoord;
	vec3 currNorm, occlNorm;
	float currDepth, occlDepth, sampleDepth;
	vec3 wCurrCoord, wSampleCoord, wOcclCoord;
	vec3 sample, signedSample, bounceDir, currToOccl;
	vec3 bounceColor = vec3(0.0);
	float sampleDist;
	float receiverCoef, senderCoef, multiplier, notShadow;
	vec3 occlColor, lightColor;

	//get normal and depth of current fragment
	currDepth = getDepth(currCoord);
	currNorm = getNormal(currCoord);
	//compute its world coordinates and offset it in the direction of the normal
	wCurrCoord = getWorldPos(currCoord,currDepth);
	wCurrCoord += 0.05*currNorm;
	
	//set pseudo-random number as sample offset
	float sampleOffset = gl_FragCoord.x*gl_FragCoord.y;;
	float rand;
	multiplier = 2.0*rmax; //rmax*rmax;

	for(int i=0;i<sampleCountB;i++)
	{	
		//set pseudo-random lenght of sample
		rand = rmax*( float(i+1)/float(sampleCountB) );
		//get sample direction and multiply it by random length
		sample = rand*getSample(i,sampleOffset);
		//transform sample to lie in the hemisphere of current fragment normal
		signedSample = sign(dot(sample,currNorm))*sample;

		//get sample in world coordinates
		wSampleCoord = wCurrCoord + signedSample;			
		sampleDepth = viewDepth(wSampleCoord);

		//determine which fragment (texet) coresponds to world coordinates of the sample
		occlCoord = getTexturePos(wSampleCoord, sampleDepth);
		
		//compute bounce only if sample is not in shadow
		notShadow = 1.0-step(1.0, getDpsMask(occlCoord));

			//get normal and depth of corresponding fragmet 
			occlNorm = getNormal(occlCoord);
			occlDepth = getDepth(occlCoord);

			//get world position of corresponding fragmet 
			wOcclCoord = getWorldPos(occlCoord,occlDepth);
			//create vector from current fragmet to sample fragmet
			currToOccl = wOcclCoord - wCurrCoord;

			//sample fragmet distance to current fragment
			sampleDist = max(length(currToOccl), 1.0);
			//light transport direction
			bounceDir = normalize(currToOccl);
			//angle coeficients between reciever/sender fragment and light transport direction
			receiverCoef = max( dot(currNorm,bounceDir), 0.0);
			senderCoef = max( dot(occlNorm,-bounceDir), 0.0);
			//color of light being bounced
			lightColor = getLight(occlCoord);
			//color of fragment bouncing the light
			occlColor = getColor(occlCoord);
			
			//final color of incoming light (from single sample direction)
			bounceColor += notShadow * lightColor*occlColor*receiverCoef*senderCoef/(sampleDist*sampleDist);
	}
	//normalize to number of samples
	return multiplier*bounceColor / float(sampleCountB);
}

void main( void )
{
	setRealWorld();
	vec2 fragCoord = textureCoord(gl_FragCoord.xy);
	
	vec3 bounceColor = vec3(0.0);
	if (sampleCountB>0){
		bounceColor = getBounce(fragCoord);
	}
	
	gl_FragData[0] = vec4(bounceColor,1.0);
}
