//textures
uniform sampler2D ndtexture;
uniform sampler2D ndtextureback;
uniform sampler2D ndtextureB;
uniform sampler2D ndtexturebackB;
uniform sampler2D positiontexture;
uniform sampler2D envMap;
uniform sampler2D sampleTexture;
uniform sampler2D noiseTexture;

//viewport & lights
uniform vec2 viewport;
uniform vec2 giTextureSize;
uniform mat4 farPlaneVertices;
uniform vec3 cameraPos;
uniform mat4 viewMatrix;
uniform int maximum;

//ambient occlusion
uniform int sampleCount;
uniform int samplesPerDir;
uniform float rmax;
uniform bool aoOnlyInShadows;
uniform bool depthPeeling;

//depth peeling shadows
uniform float dpsRmax;
uniform int sampleNo;

mat3 toTextureMatrix;
mat3 toTextureMatrixInverse;

float near, far;
vec3 fnright, fnup, fneye;
float fwidth, fheight;

////////////////////////////////////////////////////////////////////////////////////

//transforms fragment coordinates <x,y> to texture coordinates <0,1>
vec2 textureCoord(const vec2 coord){
	return coord/giTextureSize;
}

//returns fragment normal
vec3 getNormal(in vec2 coord){
	return texture2D( ndtexture, coord ).xyz;
}

//returns fragment depth
float getDepth(in vec2 coord){
	return texture2D( ndtexture, coord ).a;
}

//returns back fragment normal
float getBackDepth(in vec2 coord){
	return texture2D( ndtextureback, coord ).a;
}

//returns fragment normal with Bilinear interpolation
vec3 getNormalB(in vec2 coord){
	return texture2D( ndtextureB, coord ).xyz;
}

//returns fragment depth with Bilinear interpolation
float getDepthB(in vec2 coord){
	return texture2D( ndtextureB, coord ).a;
}

//returns back fragment normal with Bilinear interpolation
float getBackDepthB(in vec2 coord){
	return texture2D( ndtexturebackB, coord ).a;
}

//returns random color
vec3 getNoise(in vec2 coord){	
	return texture2D( noiseTexture, coord ).xyz;
}

//returns sample from texture
vec3 getSample(in int i, in float j, in vec2 curr)
{	
	vec2 coord;
	//x - number of sample
	//y - number of set of samples
	coord.x = float(i)/float(sampleCount);
	coord.y = j/128.0 + getNoise(curr).x;
	return 2.0*texture2D( sampleTexture, coord ).xyz - 1.0;
}

//computes inverse matrix 3x3
mat3 inverseMatrix(in mat3 m)
{
	float det00, det01, det02, det10, det11, det12, det20, det21, det22;
	float detA = determinant(m);

	det00 = determinant(mat2( m[1].y, m[2].y, m[1].z, m[2].z ));
	det01 = determinant(mat2( m[0].z, m[2].z, m[0].y, m[2].y ));
	det02 = determinant(mat2( m[0].y, m[1].y, m[0].z, m[1].z ));

	det10 = determinant(mat2( m[1].z, m[2].z, m[1].x, m[2].x ));
	det11 = determinant(mat2( m[0].x, m[2].x, m[0].z, m[2].z ));
	det12 = determinant(mat2( m[0].z, m[1].z, m[0].x, m[1].x ));

	det20 = determinant(mat2( m[1].x, m[2].x, m[1].y, m[2].y ));
	det21 = determinant(mat2( m[0].y, m[2].y, m[0].x, m[2].x ));
	det22 = determinant(mat2( m[0].x, m[1].x, m[0].y, m[1].y ));

	return mat3(det00, det01, det02, det10, det11, det12, det20, det21, det22)/detA;
}

//sets up global view variables
void setRealWorld()
{
	float znear = gl_ProjectionMatrix[2].z;
	float zfar = gl_ProjectionMatrix[3].z;
	near = zfar / (znear - 1.0);
	far = zfar / (znear + 1.0);

	vec3 fbl, fbr, fur, ful, fcenter;
	vec3 fright, fup, feye;

	//far clipping plane vertices
	//b-bottom, u-upper, r-right, l-left
	fbl = farPlaneVertices[0].xyz;
	fbr = farPlaneVertices[1].xyz;
	fur = farPlaneVertices[2].xyz;
	ful = farPlaneVertices[3].xyz;

	//center point of far clipping plane
	fcenter = (fbl+fbr+ful+fur)*0.25;

	fright = fbr - fbl;
	fup = ful - fbl;
	feye = fcenter-cameraPos;
	fwidth = length(fright);
	fheight = length(fup);

	fnright = normalize(fright);
	fnup = normalize(fup);
	fneye = normalize(feye);

	//matrices for transforming between world and texture coordinates
	toTextureMatrix = transpose(mat3(fnright, fnup, fneye));
	toTextureMatrixInverse = inverseMatrix( toTextureMatrix );
}

//computes view depth of 3d point
float viewDepth(in vec3 pos)
{	
	//get view vector
	vec3 viewVec = pos-cameraPos;
	//get depth as lenght of the view vector
	float depth = length(viewVec);
	//compute correction angle
	float angle = dot(normalize(viewVec),fneye);

	//correct depth and transform to <0,1>
	return depth*angle/far;
}

//computes position of a point in world space from its window coordinates and depth
vec3 getWorldPos(const vec2 coord, const float depth)
{
	//compute coordinates in window
	vec3 tcoord = vec3( (coord-0.5)*vec2(fwidth,fheight), far);
	//get view vector by transforming coordinates to world space
	vec3 viewVec = toTextureMatrixInverse*tcoord;
	
	//add point vector and depth to the position of a camera
	vec3 worldPos = cameraPos + depth * viewVec;
    return worldPos;
	
	//alternative
	//return texture2D( positiontexture, coord ).xyz;
}

//computes texture coordinates of a point from its worldSpace coordinates and depth
vec2 getTexturePos(const vec3 wpos, const float depth)
{
	//get view vector
	vec3 viewVec = ( wpos - cameraPos ) / depth;
	//transform it to window coordinats
	vec3 coord = toTextureMatrix*viewVec;
	
	//clipping plane to texture coordinates
	return coord.xy / vec2(fwidth,fheight) + 0.5;

	//alternative
	/*
	vec4 position = gl_ProjectionMatrix * viewMatrix * vec4(wpos,1.0);
	vec4 nPosition = position / position.w;
	return (nPosition.xy+1.0)*0.5;*/
}
////////////////////////////////////////////////////////////////////////////////////

//returns environmental color from texture according to sample direction
//after optimization
vec3 getEnv(in vec3 envDir)
{	
	vec2 coord;
	coord.s = envDir.x*0.25;
	coord.t = envDir.y*0.5+0.5;	

	if (envDir.z >= 0.0){
		coord.s = coord.s + 0.5;
	}else{
		coord.s = 1.0 - coord.s;
	}

	return texture2D( envMap, coord ).xyz;
}

//computes directional ambient occlusion and environmental color of current fragment
vec4 getOcclusion(in vec2 currCoord, in float dpsMask)
{	
	vec2 occlCoord;
	vec3 currNorm;
	float currDepth, occlDepth, occlBackDepth;
	vec3 wCurrCoord, wSampleCoord;
	float sampleDepth, occlTest, occlBackTest;
	vec3 envDir, sampleDir, signedSampleDir, sample, signedSample;
	vec3 envColor = vec3(0.0);
	float occlusion, occlFinalAdd;
	float occlusionFinal = 0.0;
	float envCoef = 1.0;
	
	//get normal and depth of current fragment
	currDepth = getDepthB(currCoord);
	currNorm = getNormalB(currCoord);
	//compute its world coordinates and offset it in the direction of the normal
	wCurrCoord = getWorldPos(currCoord,currDepth);
	wCurrCoord += 0.05*currNorm;
	
	//set pseudo-random number as sample offset
	float sampleOffset = gl_FragCoord.x*gl_FragCoord.y;
	float rand;
	float distanceTerm;

	//compute ambient occlusion only in shadow
	if (dpsMask>0.0 || !aoOnlyInShadows)
	{	
		//sample 3d neighbourhood of current fragment
		for(int i=0;i<sampleCount;i++)
		{	
			//get sample direction
			sampleDir = getSample(i,sampleOffset,currCoord);
			//transform sample to lie in the hemisphere of current fragment's normal
			signedSampleDir = sign(dot(sampleDir,currNorm))*sampleDir;

			occlusion = 0.0;
			//sample each direction
			for(int j=1;j<(samplesPerDir+1);j++)
			{	
				//set pseudo-random lenght of sample
				rand = rmax*( (float(i+1)*float(j)) / (float(sampleCount)*float(samplesPerDir)) );

				//multiply sample by random length
				signedSample = rand*signedSampleDir;
				
				//get sample in world coordinates
				wSampleCoord = wCurrCoord + signedSample;			
				sampleDepth = viewDepth(wSampleCoord);

				//determine which fragment (texel) coresponds to world coordinates of the sample
				occlCoord = getTexturePos(wSampleCoord, sampleDepth);

				//get depth of front faces
				occlDepth = getDepthB(occlCoord);

				//determine position of sample to geometry of the scene
				//front test:	sampleDepth < occlDepth		=> sample is in front of geometry	= 1
				//back test:	sampleDepth > occlBackDepth => sample is behind geometry		= 1
				//otherwise:								   sample is inside geometry		= 0
				
				if(depthPeeling)
				{
					occlBackDepth = getBackDepthB(occlCoord);
					occlTest = step(sampleDepth,occlDepth);				
					occlBackTest = step(occlBackDepth,sampleDepth);
				}else{
					distanceTerm = step(abs(currDepth-occlDepth),rmax/far);
					occlTest = 1.0-step(occlDepth,sampleDepth)*distanceTerm;	
				}

				//1: sample is 'visible'
				//0: sample is inside geometry, i.e. 'invisible' or 'occluder'
				occlusion += occlTest + occlBackTest;		
			}		
			//test if any of samples made per one direction was 'invisible'
			occlFinalAdd = step(float(samplesPerDir), occlusion);
			occlusionFinal += occlFinalAdd;

			//compute environmental color comming from sample direction
			envDir = signedSampleDir;
			envCoef = max(0.0, dot(currNorm, envDir));
			envColor += getEnv(envDir)*envCoef*occlFinalAdd;
		}
	}else{
		//don't compute ambient occlusion, fragment is not in shadow
		//compute only environmental color
		for(int i=0;i<sampleCount;i++)
		{	
				sample = getSample(i,sampleOffset,currCoord);
				signedSample = sign(dot(sample,currNorm))*sample;
				
				envDir = normalize(signedSample);
				envCoef = max(0.0, dot(currNorm, envDir));
				envColor += getEnv(envDir)*envCoef;
		}

		//set ambient occlusion to '1.0' (after normalization)
		occlusionFinal = float(sampleCount);
	}
	//normalize to number of samples
	return vec4(envColor,occlusionFinal) / float(sampleCount) ;
}

//computes colored shadows in image (screen) space
vec4 getDepthPeelingShadows(in vec2 currCoord)
{
	vec2 occlCoord;
	vec3 currNorm;
	float currDepth, occlDepth, occlBackDepth;
	vec3 wCurrCoord, wSampleCoord;
	float sampleDepth, occlTest, occlBackTest;
	float lightCoef, occlusion, occlFinal;
	vec3 pointToLight, lightDir, lightNorm;
	vec2 currCoordR, occlCoordR;

	//compensation for inaccuracy of float computing
	float round = 100000.0;
	currCoordR = floor(currCoord*round)/round;

	//get normal and depth of current fragment
	currDepth = getDepth(currCoordR);
	currNorm = getNormal(currCoordR);
	//compute its world coordinates and offset it in the direction of the normal
	wCurrCoord = getWorldPos(currCoordR,currDepth);
	wCurrCoord += 0.05*currNorm;
		
	//initialize color of shadows
	vec3 lightColor = vec3(0.0);
	float lightMask = 0.0;
	float meanSum = 0.0;

	for(int j=0; j<maximum; j++)
	{
		//get inverse light vector
		pointToLight = vec4(gl_ModelViewMatrixInverse*gl_LightSource[j].position).xyz;
		pointToLight -= wCurrCoord * gl_LightSource[j].position.w;
		
		lightNorm = normalize(pointToLight);
		lightCoef = dot(lightNorm,currNorm);
		//compute sample length
		lightDir = dpsRmax*lightNorm / float(sampleNo);

		occlusion = 0.0;
		//compute color only if fragment is visible to light source
		if ( lightCoef > 0.0 )
		{		
			for(int i=1;i<(sampleNo+1);i++)
			{	
				//get world position of sample
				wSampleCoord = wCurrCoord + lightDir * float(i);
				sampleDepth = viewDepth(wSampleCoord);

				//determine which fragment (texet) coresponds to world coordinates of the sample
				occlCoord = getTexturePos(wSampleCoord, sampleDepth);

				//get depths of front and back faces
				occlDepth = getDepth(occlCoord);
				occlBackDepth = getBackDepth(occlCoord);

				//determine position of sample to geometry of the scene
				occlTest = step(sampleDepth,occlDepth);				
				occlBackTest = step(occlBackDepth,sampleDepth);

				//test if sample is 'visible' or inside geometry
				occlusion += occlTest+occlBackTest;
			}

			//compute light coeficient for shadow strength
			float lightMean = (gl_LightSource[j].diffuse.x + gl_LightSource[j].diffuse.y + gl_LightSource[j].diffuse.z) / 3.0;
			meanSum += lightMean;

			//add incoming light from light source, if not in shadow
			occlFinal = step(float(sampleNo), occlusion);
			lightColor += gl_LightSource[j].diffuse.xyz*lightCoef*occlFinal;
			lightMask += occlFinal*lightMean;
		}
	}
	//return shadow color and shadow mask
	return vec4(lightColor, max(1.0-lightMask/meanSum,0.0));
}

void main( void )
{
	//set global parapeters and transform texture coordinets to <0,1>
	setRealWorld();
	vec2 fragCoord = textureCoord(gl_FragCoord.xy);

	vec4 fragDps = vec4(0.0);
	vec4 fragDO = vec4(vec3(0.0),1.0);

	//compute shadows only if there are any lights
	float dpsMask;
	if (sampleNo>0 && maximum>0){
		fragDps = getDepthPeelingShadows(fragCoord);
		dpsMask = fragDps.w;
	}else{
		//everithing is in shadow
		dpsMask = 1.0;
	}

	//compute directional occlusion
	if (sampleCount>0){
		fragDO = getOcclusion(fragCoord,dpsMask);
	}
	
	gl_FragData[0] = fragDO;
	gl_FragData[1] = fragDps;
}