/*
 * VRECKO - Main fragment shader
 * ----------------------
 *
 *	Contains main functions, with phong lighting computation
 *	Adding another function is possible
 *	
 *	Martin Bezdeka, <mbezdeka@mail.muni.cz>
 *
 *	Last updated: 3. 4. 2010
 *
 */
varying vec3 normal, vVertex;
varying vec2 texCoord;

uniform sampler2D osgShadow_baseTexture;

uniform int maximum;
uniform bool shadowsEnabled;
uniform bool colorMaterial;
uniform bool twoSided;

void directionalLight(in int i, in vec3 normal, inout vec4 ambient, inout vec4 diffuse, inout vec4 specular);
void pointLight(in int i, in vec3 normal, inout vec4 ambient, inout vec4 diffuse, inout vec4 specular);
void spotLight(in int i, in vec3 normal, inout vec4 ambient, inout vec4 diffuse, inout vec4 specular);

void shadowCompute(inout vec4 color);

void main(void)
{	
	vec4 ambient = vec4(0.0);
	vec4 diffuse = vec4(0.0);
	vec4 specular = vec4(0.0);
	
	vec4 sumOfLights = vec4(0.0);
	
	for(int i=0; i<maximum; i++)
	{
		if(gl_LightSource[i].position.w == 0.0)
			directionalLight(i,normal,ambient,diffuse,specular);
		else if(gl_LightSource[i].spotCutoff == 180.0)
			pointLight(i,normal,ambient,diffuse,specular);
		else
			spotLight(i,normal,ambient,diffuse,specular);
			
	}
	
	if(colorMaterial)
		sumOfLights = ambient * gl_Color + diffuse * gl_Color + specular * gl_FrontMaterial.specular;
	else
		sumOfLights = ambient * gl_FrontMaterial.ambient + 
					   diffuse * gl_FrontMaterial.diffuse +
					   specular * gl_FrontMaterial.specular;
	
	//Final light equation
	vec4 finalColor =  gl_LightModel.ambient * gl_FrontMaterial.ambient + gl_FrontMaterial.emission + sumOfLights;
	
	//If two-sided lighting enabled, compute lighting again for back side with reversed normals
	if(twoSided)
	{
		ambient = vec4(0.0);
		diffuse = vec4(0.0);
		specular = vec4(0.0);
	
		for(int i=0; i<maximum; i++)
		{
			if(gl_LightSource[i].position.w == 0.0)
				directionalLight(i,-normal,ambient,diffuse,specular);
			else if(gl_LightSource[i].spotCutoff == 180.0)
				pointLight(i,-normal,ambient,diffuse,specular);
			else
				spotLight(i,-normal,ambient,diffuse,specular);
		}
		
		if(colorMaterial)
			sumOfLights = ambient * gl_Color + diffuse * gl_Color + specular * gl_BackMaterial.specular;
		else
			sumOfLights = ambient * gl_BackMaterial.ambient + 
						  diffuse * gl_BackMaterial.diffuse +
						  specular * gl_BackMaterial.specular;
		
		//Finalni rovnice osvetleni pridani yadniho osvetleni
		finalColor +=  gl_LightModel.ambient * gl_BackMaterial.ambient + gl_BackMaterial.emission + sumOfLights;
	}


	//Adding texture
	vec4 texColor = texture2D(osgShadow_baseTexture, texCoord);
	if(texColor != vec4(0))
		finalColor *= texColor;
	
	//----------------------------------------------------------------------
	//				ADDITIONAL COMPUTATIONS	- add your own function
	//----------------------------------------------------------------------
	
	//ShadowMap and SoftShadowMap computation
	if(shadowsEnabled)	
		shadowCompute(finalColor);
		
	//----------------------------------------------------------------------
	//				END OF ADDITIONAL COMPUTATIONS
	//----------------------------------------------------------------------

	gl_FragColor = vec4(finalColor.xyz, gl_Color.a*gl_FrontMaterial.diffuse.a);
}



//////////////////////////////////////////////////////////
//Functions

void directionalLight(in int i, in vec3 normal, inout vec4 ambient, inout vec4 diffuse, inout vec4 specular)
{
	vec3 l = gl_LightSource[i].position.xyz;
	
	vec3 N = normalize(normal);
	vec3 L = normalize(l);
		
	float nDotL = max(dot(L,N),0.0);
	
	ambient += gl_LightSource[i].ambient;
	diffuse += gl_LightSource[i].diffuse * nDotL;
			
	if(gl_FrontMaterial.shininess != 0.0)	//Pokud neni nastaveny parametr lesklosti, nebude odlesk pocitan
	{
		float pf = max(dot(N,gl_LightSource[i].halfVector.xyz),0.0);
		
		specular += gl_LightSource[i].specular * pow( pf,gl_FrontMaterial.shininess );
	}
}

void pointLight(in int i, in vec3 normal, inout vec4 ambient, inout vec4 diffuse, inout vec4 specular)
{
		//Vektor od vrcholu do pozice svetla
		vec3 l = gl_LightSource[i].position.xyz - vVertex;
		
		//velikost vektoru l
		float d = length(l);
		
		//normalizace vektoru
		vec3 N = normalize(normal);
		vec3 L = normalize(l);
		
		float nDotL = max(dot(L,N),0.0);
		
					
		float att = 1.0 / (gl_LightSource[i].constantAttenuation +
						  (gl_LightSource[i].linearAttenuation*d) +
						  (gl_LightSource[i].quadraticAttenuation*d*d) );
	
		///////////////////////////////////////////////////////////////////////////////////							  
		//Vypocet ambientni slozky
		
		ambient += gl_LightSource[i].ambient * att;
		
		if(nDotL > 0.0)
		{		
			///////////////////////////////////////////////////////////////////////////////////
			//Vypocet difuzni slozky			
			
			diffuse += gl_LightSource[i].diffuse * nDotL * att;
			
			///////////////////////////////////////////////////////////////////////////////////
			//Vypocet spekularni slozky
	
			if(gl_FrontMaterial.shininess != 0.0)
			{			
				vec3 E = normalize(-vVertex);
				vec3 S = reflect(-L,N);
					
				float pf = max(dot(S,E),0.0);
				
				specular += gl_LightSource[i].specular * pow( pf,gl_FrontMaterial.shininess ) * att;
			}
		} 
}

void spotLight(in int i, in vec3 normal, inout vec4 ambient, inout vec4 diffuse, inout vec4 specular)
{
		//Vektor od vrcholu do pozice svetla
		vec3 l = gl_LightSource[i].position.xyz - vVertex;
		
		//velikost vektoru l
		float d = length(l);
		
		//normalizace vektoru
		vec3 N = normalize(normal);
		vec3 L = normalize(l);
		
		float nDotL = max(dot(L,N),0.0);
		
		///////////////////////////////////////////////////////////////////////////////////
		//vypocet utlumu
			
		float att = 1.0 / (gl_LightSource[i].constantAttenuation +
						  (gl_LightSource[i].linearAttenuation*d) +
						  (gl_LightSource[i].quadraticAttenuation*d*d) );

		///////////////////////////////////////////////////////////////////////////////////							  
		//Vypocet ambientni slozky
	
		if(nDotL > 0.0)
		{			
			
			///////////////////////////////////////////////////////////////////////////////////
			//Vypocet spekularni slozky
			
			float pf = 0;
			if(gl_FrontMaterial.shininess != 0.0)
			{
				vec3 E = normalize(-vVertex);
				vec3 S = reflect(-L,N);
					
				pf = max(dot(S,E),0.0);
			}

			///////////////////////////////////////////////////////////////////////////////////
			//Spot efekt - Kuzelove svetlo
			float spot = 1.0;
			
			float spotAngle = max(dot(-L,normalize(gl_LightSource[i].spotDirection)),0.0);
			if(spotAngle > gl_LightSource[i].spotCosCutoff)
				spot = pow( spotAngle ,gl_LightSource[i].spotExponent);
			else
				spot = 0.0;
			
			att *= spot;
			
			///////////////////////////////////////////////////////////////////////////////////
			//Vypocet difuzni slozky			
			
			diffuse += gl_LightSource[i].diffuse * nDotL * att;
			if(gl_FrontMaterial.shininess != 0.0)
				specular += gl_LightSource[i].specular * pow( pf,gl_FrontMaterial.shininess ) * att;
		}
		
		ambient += gl_LightSource[i].ambient * att;
}