uniform sampler2D osgShadow_baseTexture;
uniform int maximum;

varying vec3 normal, wNormal;
varying vec3 vertex;
varying vec4 wVertex;
varying vec2 texCoord;

//Lights
void getLight(in vec2 currCoord, inout vec4 diffuse, inout vec4 specular);
void directionalLight(in int i, in vec3 normal, inout vec4 diffuse, inout vec4 specular);
void pointLight(in int i, in vec3 normal, inout vec4 diffuse, inout vec4 specular);
void spotLight(in int i, in vec3 normal, inout vec4 diffuse, inout vec4 specular);

void main( void )
{
	float znear = gl_ProjectionMatrix[2].z;
	float zfar = gl_ProjectionMatrix[3].z;
	float near = zfar / (znear - 1.0);
	float far = zfar / (znear + 1.0);

	float depth = 1.0 / (gl_FragCoord.w * far);

	vec4 textureColor = texture2D(osgShadow_baseTexture, texCoord);
	if(textureColor == vec4(0))
		textureColor = gl_FrontMaterial.diffuse;
		
	vec4 lightDiffColor, lightSpecColor;
	getLight(texCoord,lightDiffColor, lightSpecColor);
	
	gl_FragData[0] = vec4(normalize(wNormal), depth);
	gl_FragData[1] = vec4(normalize(wNormal), depth);

	//for bad normal interpolation in models
	//vec3 finalNormal = cross( dFdx(wVertex.xyz), dFdy(wVertex.xyz) );
	//gl_FragData[0] = vec4(normalize(finalNormal), depth);
	//gl_FragData[1] = vec4(normalize(finalNormal), depth);
	
	gl_FragData[2] = wVertex;
	gl_FragData[3] = textureColor;
	gl_FragData[4] = lightDiffColor;
	gl_FragData[5] = lightSpecColor;
}

///////////////////////////////////////////////////////////////////////////////
void getLight(in vec2 currCoord, inout vec4 diffuse, inout vec4 specular)
{
	diffuse = vec4(0.0);
	specular = vec4(0.0);
		
	for(int i=0; i<maximum; i++)
	{
		if(gl_LightSource[i].position.w == 0.0)
			directionalLight(i,normal,diffuse,specular);
		else if(gl_LightSource[i].spotCutoff == 180.0)
			pointLight(i,normal,diffuse,specular);
		else
			spotLight(i,normal,diffuse,specular);
	}
}

void directionalLight(in int i, in vec3 normal, inout vec4 diffuse, inout vec4 specular)
{
	//Vektor od vrcholu do pozice svetla
	vec3 l = gl_LightSource[i].position.xyz;
	
	vec3 N = normalize(normal);
	vec3 L = normalize(l);
		
	float nDotL = max(dot(L,N),0.0);
	
	diffuse += gl_LightSource[i].diffuse * nDotL;
			
//	if(gl_FrontMaterial.shininess != 0.0)	//Pokud neni nastaveny parametr lesklosti, nebude odlesk pocitan
//	{
		float pf = max(dot(N,gl_LightSource[i].halfVector.xyz),0.0);
		
		vec4 lsSpecular = vec4(1.0);
		float fmShininess = 50.0;

		//specular += gl_LightSource[i].specular * pow( pf,gl_FrontMaterial.shininess );
		specular += lsSpecular * pow( pf,fmShininess );
//	}
}

void pointLight(in int i, in vec3 normal, inout vec4 diffuse, inout vec4 specular)
{
		//Vektor od vrcholu do pozice svetla
		vec3 l = gl_LightSource[i].position.xyz - vertex;
		
		//velikost vektoru l
		float d = length(l);
		
		//normalizace vektoru
		vec3 N = normalize(normal);
		vec3 L = normalize(l);
		
		float nDotL = max(dot(L,N),0.0);
		
					
		float att = 1.0 / (gl_LightSource[i].constantAttenuation +
						  (gl_LightSource[i].linearAttenuation*d) +
						  (gl_LightSource[i].quadraticAttenuation*d*d) );
	
		///////////////////////////////////////////////////////////////////////////////////							  
		
		if(nDotL > 0.0)
		{		
			///////////////////////////////////////////////////////////////////////////////////
			//Vypocet difuzni slozky			
			
			diffuse += gl_LightSource[i].diffuse * nDotL * att;
			
			///////////////////////////////////////////////////////////////////////////////////
			//Vypocet spekularni slozky
	
			if(gl_FrontMaterial.shininess != 0.0)
			{			
				vec3 E = normalize(-vertex);
				vec3 S = reflect(-L,N);
					
				float pf = max(dot(S,E),0.0);
				
				specular += gl_LightSource[i].specular * pow( pf,gl_FrontMaterial.shininess ) * att;
			}
		} 
}

void spotLight(in int i, in vec3 normal, inout vec4 diffuse, inout vec4 specular)
{
		//Vektor od vrcholu do pozice svetla
		vec3 l = gl_LightSource[i].position.xyz - vertex;
		
		//velikost vektoru l
		float d = length(l);
		
		//normalizace vektoru
		vec3 N = normalize(normal);
		vec3 L = normalize(l);
		
		float nDotL = max(dot(L,N),0.0);
		
		///////////////////////////////////////////////////////////////////////////////////
		//vypocet utlumu
			
		float att = 1.0 / (gl_LightSource[i].constantAttenuation +
						  (gl_LightSource[i].linearAttenuation*d) +
						  (gl_LightSource[i].quadraticAttenuation*d*d) );

		///////////////////////////////////////////////////////////////////////////////////							  
	
		if(nDotL > 0.0)
		{			
			
			///////////////////////////////////////////////////////////////////////////////////
			//Vypocet spekularni slozky
			
			float pf = 0.0;
			if(gl_FrontMaterial.shininess != 0.0)
			{
				vec3 E = normalize(-vertex);
				vec3 S = reflect(-L,N);
					
				pf = max(dot(S,E),0.0);
			}

			///////////////////////////////////////////////////////////////////////////////////
			//Spot efekt - Kuzelove svetlo
			float spot = 1.0;
			
			float spotAngle = max(dot(-L,normalize(gl_LightSource[i].spotDirection)),0.0);
			if(spotAngle > gl_LightSource[i].spotCosCutoff)
				spot = pow( spotAngle ,gl_LightSource[i].spotExponent);
			else
				spot = 0.0;
			
			att *= spot;
			
			///////////////////////////////////////////////////////////////////////////////////
			//Vypocet difuzni slozky			
			
			diffuse += gl_LightSource[i].diffuse * nDotL * att;
			if(gl_FrontMaterial.shininess != 0.0)
				specular += gl_LightSource[i].specular * pow( pf,gl_FrontMaterial.shininess ) * att;
		}
}